/*
   TcpServer Class for C++ by Viktor "AlgorithMan" Engelmann

   This source code is provided under the terms of the
   GNU general public license version 3. See gpl3.txt for details.

   thanks to
      http://www.zotteljedi.de/
      http://www.linuxhowtos.org/C_C++/socket.htm
*/

#ifdef DEBUG
     #define DBG(X) std::cerr << X << std::endl;
#else
     #define DBG(X)
#endif

#include "tcpserver.h"
using namespace std;
using namespace AlgorithMan;

void* Client_MonitorThreadProc(void* c)
{
    Client* C = (Client*) c;
    try
    {
        C->MonitorThreadProc();
    }
    catch(char const* msg)
    {
        std::cerr << msg << std::endl;
    }
    delete C;
    return nullptr;
}

void Client::MonitorThreadProc()
{
    const int max_chars = 255;
    char buffer[max_chars+1];
    int n;

    isConnected = true;
    server->OnConnect(this);
    while (true)
    {
        n = read(clientSocket, &buffer, max_chars);
        if (n <= 0)
            break;
        std::string message(buffer, n);
        server->OnReceive(this, message);
    }

    server->NotifyDisconnect(this);
    close(clientSocket);
    isConnected = false;
}


Client::Client(TcpServer* server)
{
    this->isConnected = false;
    this->server = server;
}

bool Client::Accept(int listeningSocket)
{
    socklen_t clilen = sizeof(client_address);
    clientSocket = accept(listeningSocket, (struct sockaddr *) &client_address, &clilen);
    if (clientSocket < 0)
        return false;
    DBG("client connected, socket-handle " << clientSocket)

    if (pthread_create(&monitorThread, NULL, Client_MonitorThreadProc, this) != 0)
        throw "Client::Accept error creating thread";
    if (pthread_detach(monitorThread) != 0)
        throw "Client::Accept error detaching thread";

    return true;
}

void Client::Disconnect()
{
    if(isConnected)
    {
        shutdown(clientSocket, SHUT_RD);
        close(clientSocket);
    }
}

void Client::Send(const std::string& message)
{
    if(!isConnected)
        throw "Sending on an unconnected socket";
    size_t n = write(clientSocket, &message[0], message.size());
    if(n < message.size())
        throw "Sending message failed";
}

std::string Client::RemoteIP()
{
    if (!isConnected)
        throw "Can\'t determine remote IP from an unconnected socket";
    return inet_ntoa(client_address.sin_addr);
}

// -------------------------------------------------------------------------------------------

void* TcpServer_ListenThreadProc(void* s)
{
    TcpServer* S = (TcpServer*) s;
    try
    {
        S->ListenThreadProc();
    }
    catch(char const* msg)
    {
        std::cerr << msg << std::endl;
    }
    return nullptr;
}

void TcpServer::ListenThreadProc()
{
    DBG("creating socket on port " << this->listeningPort << "...")
    listeningSocket = socket(AF_INET, SOCK_STREAM, 0);
    if (listeningSocket < 0)
        throw "could not create socket";

    int enable = 1;
    if (setsockopt(listeningSocket, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(int)) < 0)
        throw "setsockopt failed";

    bzero((char *) &server_address, sizeof(server_address));
    server_address.sin_family = AF_INET;
    server_address.sin_port = htons(this->listeningPort);
    server_address.sin_addr.s_addr = INADDR_ANY;
    DBG("binding...")
    if (bind(listeningSocket, (struct sockaddr *) &server_address, sizeof(server_address)) < 0)
        return;

    DBG("listening...")
    listen(listeningSocket, 5);
    isListening = true;
    while (isListening)
    {
        Client* client = InstantiateClient();
        if(!client->Accept(listeningSocket))
        {
            delete client;
            break;
        }

        pthread_mutex_lock(&mutex);
        clients.push_back(client);
        pthread_mutex_unlock(&mutex);
    }
    isListening = false;
}

TcpServer::TcpServer()
{
    isListening = false;
    listeningPort = -1;
    pthread_mutex_init(&mutex, nullptr);
}

TcpServer::~TcpServer()
{
    Stop();
    pthread_mutex_destroy(&mutex);
}

void TcpServer::Start(int port)
{
    if (isListening)
        return;
    this->listeningPort = port;
    if (pthread_create(&listenThread, NULL, TcpServer_ListenThreadProc, this) != 0)
        throw "Could not create thread";
}

void TcpServer::Stop()
{
    if (!isListening)
        return;

    isListening = false;
    DBG("closing listeningSocket " << listeningSocket)
    shutdown(listeningSocket, SHUT_RD);
    close(listeningSocket);
    DBG("waiting for main thread to terminate...")
    pthread_join(listenThread, NULL);

    DBG("disconnecting all clients...")
    for(Client* c : clients)
        c->Disconnect(); // terminates the respective threads
                         // the threads delete the objects at the end
    clients.clear();
}

void TcpServer::OnConnect(Client* client)
{
}

void TcpServer::NotifyDisconnect(Client* client)
{
    this->OnDisconnect(client);

    // otherwise, we are shutting down and this would deadlock
    // against the loop that closes all open sockets
    if(isListening)
    {
        pthread_mutex_lock(&mutex);
        auto it = std::find(clients.begin(), clients.end(), client);
        if (it != clients.end())
            clients.erase(it);
        pthread_mutex_unlock(&mutex);
    }
}

void TcpServer::OnDisconnect(Client* sender)
{
}

void TcpServer::OnReceive(Client* sender, const std::string& message)
{
}

void TcpServer::Send(Client* client, const std::string& message)
{
    client->Send(message);
}

void TcpServer::Broadcast(const std::string& message)
{
    pthread_mutex_lock(&mutex);
    for(Client* c : clients)
        Send(c, message);
    pthread_mutex_unlock(&mutex);
}
